//      
// COM-  1-Wire .

#include <sstream>

#include <iostream>
#include <iomanip>
#include <vector>

#if _MSC_VER < 1300
typedef unsigned __int64 uint64_t;
#else
#include <stddef.h>
#endif

#include "com_port.h"

//     com- (  1: COM1, COM2, ...).
#define COM_NUM 2

//     ( 
//   MCU).
#define COM_BAUD_RATE 2400

//        1-Wire.

// ROM-
#define CMD_SEARCH_ROM          0xF0u
#define CMD_READ_ROM            0x33u
#define CMD_MATCH_ROM           0x55u
#define CMD_SKIP_ROM            0xCCu
#define CMD_ALARM_SEARCH        0xECu

//  .
#define CMD_CONVERT_T           0x44u
#define CMD_WRITE_SCRATCHPAD    0x4Eu
#define CMD_READ_SCRATCHPAD     0xBEu
#define CMD_COPY_SCRATCHPAD     0x48u
#define CMD_RECALL_E2           0xB8u
#define CMD_READ_POWER_SUPPLY   0xB4u

//    CRC   :
// CRC=x**8+x**5+x**4+1
// CRC    1-Wire     ;
// CRC        ROM-;
//   scratchpad  DS18B20   CRC   .
//      .
char crc8(const char *buf, unsigned int size)
{
	unsigned char result=0;
	for(unsigned int i=0; i<size; i++)
	{
		unsigned char c=buf[i];
		for(unsigned int j=0; j<8; j++, c>>=1)
		{
			unsigned char b=(c^result)&1;
			unsigned char mask=~(b-1)&0x18;
			result^=mask;
			result>>=1;
			result|=b<<7;
		}
	}
	return result;
}


//       MCU  
//  ,   .

//      ,  
//  - ,   1   COM-,   
//    1 .

//     ,  
//    (   1-Wire)  
//  ,      .

//  ,  ,   . 
//    4 ,     
// ,   - .     4 
//      .

//   COM_NOP   0,     
//     0.      
//  ,    4    
// 4  .

//   ( .. NOP)    ,
//    .

#define COM_NOP					0x0
#define COM_RESET				0x1
#define COM_WRITE0				0x2
#define COM_WRITE1				0x3
#define COM_WRITE0_STRONG		0x4
#define COM_WRITE1_STRONG		0x5
#define COM_READ_1BIT			0x6
#define COM_READ_2BITS			0x7
#define COM_READ_3BITS			0x8
#define COM_READ_4BITS			0x9
#define COM_GRAND_STRONG_PULLUP	0xA
#define COM_NOP2				0xB
#define COM_WRITE00				0xC
#define COM_WRITE01				0xD
#define COM_WRITE10				0xE
#define COM_WRITE11				0xF

//     COM_NOP    0;
//     (   ).

//  COM_RESET    
//   1-Wire, : 0 -     
//     1-Wire; 1 -  .

//  COM_WRITE0, COM_WRITE1, COM_WRITE00, ...
//          1-Wire.
//     .

//  COM_READ_XXXX      1-Wire
//  1  4 ,       .

//    ,   
//    .
//      ,   1 :
//  4  -  COM_GRAND_STRONG_PULLUP,
//  4  -  COM_WRITE0_STRONG  COM_WRITE1_STRONG.
//       
// (,       ).

//       .

//           Com_1WireBridge.
class Com_1WireBridgeError
{
public:
};

//       COM - 1-Wire .
class Com_1WireBridge
{
private:
	Com com;
public:
	//      COM-  
	//       .
	void open(unsigned char port_num, DWORD baud_rate, BYTE byte_size=8, BYTE parity=NOPARITY, BYTE stop_bits=ONESTOPBIT)
	{
		std::basic_stringstream<_TCHAR> port_name;
		port_name<<_T("\\\\.\\COM")<<(unsigned int)port_num;
		com.open(port_name.str().c_str());
		com.config(baud_rate, byte_size, parity, stop_bits);
		COMMTIMEOUTS timeouts;
		com.get_timeouts(timeouts);
		timeouts.ReadTotalTimeoutMultiplier=10+3*12000/baud_rate;
		timeouts.ReadTotalTimeoutConstant=20;
		com.set_timeouts(timeouts);
	}

	//  COM-.
	void close()
	{
		com.close();
	}

	//       .
	//   -  .
	unsigned char send_com_cmd(unsigned char cmd)
	{
		com.write((const char*)&cmd, 1);
		char result;
		if(com.read(&result, 1)!=1)
			throw Com_1WireBridgeError();
		return result;
	}
	
	//     1-Wire;
	//  true,    .
	bool reset() {return send_com_cmd(COM_RESET)!=0;}

	//      1-Wire.
	void slot_w(bool d)
	{
		if((int)send_com_cmd(COM_WRITE0+d)!=(int)d)
			throw Com_1WireBridgeError();
	}

	//      1-Wire,
	//   .
	bool slot_r() {return send_com_cmd(COM_READ_1BIT)!=0;}

	char receive_2bits() {return send_com_cmd(COM_READ_2BITS);}

	void send_byte(char d)
	{
		unsigned char d1=d&0x03|d<<2&0x30;    //0 0 b3 b2 0 0 b1 b0
		unsigned char d2=d>>4&0x03|d>>2&0x30; //0 0 b7 b6 0 0 b5 b4
		const unsigned char CMDx2=COM_WRITE00|COM_WRITE00<<4;
		
		unsigned char r1=send_com_cmd(d1|CMDx2);
		bool succ=r1==d1;
		if(succ)
		{
			unsigned char r2=send_com_cmd(d2|CMDx2);
			succ=r2==d2;
		}
		if(!succ)
			throw Com_1WireBridgeError();
	}

	void send_rom(uint64_t rom)
	{
		char *p=(char*)&rom;
		for(unsigned int i=0; i<8; i++, p++)
			send_byte(*p);
	}

	void slot_w_and_strong_pullup(bool d)
	{
		unsigned char cmd=(d)?COM_WRITE1_STRONG:COM_WRITE0_STRONG;
		cmd|=COM_GRAND_STRONG_PULLUP;
		unsigned char r=0xFF;
		try
		{
			r=send_com_cmd(cmd);
			if(r!=d<<4)
				throw Com_1WireBridgeError();
		}
		catch(...)
		{
			//  ,     ,  - 
			//         ,  
			//       
			//  ,      .
			send_com_cmd(COM_NOP);
			throw;
		}
	}

	void send_byte_and_strong_pullup(char d)
	{
		unsigned char d1=d&0x03|d<<2&0x30;    //0 0 b3 b2 0 0 b1 b0
		unsigned char d2=d>>4&0x03; //0 0 0 0 0 0 b5 b4
		const unsigned char CMDx2=COM_WRITE00|COM_WRITE00<<4;
		
		unsigned char r1=send_com_cmd(d1|CMDx2);
		bool succ=r1==d1;
		if(succ)
		{
			unsigned char r2=send_com_cmd(d2|COM_WRITE00);
			succ=r2==d2;
		}
		
		if(succ)
		{
			d>>=6;
			slot_w(d&1);
			d>>=1;
			slot_w_and_strong_pullup(d&1);
		}

		if(!succ)
			throw Com_1WireBridgeError();
	}

	void disable_strong_pullup() {send_com_cmd(COM_NOP);}

	char receive_byte() {return send_com_cmd(COM_READ_4BITS|COM_READ_4BITS<<4);}
};

//       
// 1-Wire .		
class EnumRom
{
private:
	//   ,     ;
	//     .
	uint64_t forks;

	//    .
	int forks_count;

public:
	//  true,    CMD_ALARM_SEARCH
	// (   ),
	//  ( ) -   CMD_SEARCH_ROM ( )
	bool alarm_filter;

    //    .
    Com_1WireBridge *bridge;

	//   rom-,   
	//     next.
	uint64_t rom;

	EnumRom(Com_1WireBridge *b):forks(0), forks_count(0), alarm_filter(false),  bridge(b), rom(0) {}

	//        .
	void clear()
	{
		forks=0;
		forks_count=0;
		rom=0;
	}

	//   rom-.
	//  -1    (rom    
	// ); 0,    ROM-  1,  ROM-  
	//    .
	int next()
	{
	    if(!bridge->reset())	//    .
			return -1;
		if(alarm_filter)
			bridge->send_byte(CMD_ALARM_SEARCH);
		else
			bridge->send_byte(CMD_SEARCH_ROM);

	    int fork_pos=0;
	    for(unsigned int i=0; i<64; i++)
	    {
			char a=bridge->receive_2bits();
	        bool a1=(a&1)!=0;
	        bool a2=(a&2)!=0;

	        //     - , ,
	        //      .
	        if(a1&&a2)
                return -1;

            bool bit;

            if(a1!=a2)  //     .
                bit=a1;
            else        //    a1, a2 -    .
            {
                if(fork_pos>=forks_count)   //   .
                {
                    bit=false;
                    forks<<=1;
                    forks_count++;
                    fork_pos=forks_count;
                }
                else //        .
                {
                    int p=forks_count-fork_pos-1;
                    if(p!=0)    //    .
                        bit=(forks&(1<<p))!=0;
                    else
                    {
                        //      .
                        bit=true;
                        forks|=1;
                    }
                    fork_pos++;
                }
            }
            bridge->slot_w(bit);
			rom>>=1;
            rom|=(uint64_t)bit<<63;
	    } //end of for...
	    //          .
	    while(forks&1)
	    {
	        forks>>=1;
	        forks_count--;
	    }
	    return (forks_count>0)?1:0;
	}
};

//    uint64_t  ,   hex-.
std::ostream &operator<<(std::ostream &obj, uint64_t d)
{
	std::stringstream s;
	s<<"0x"<<std::hex<<std::setw(8)<<std::setfill('0');
	s<<(unsigned int)(d>>32&0xFFFFFFFF);
	s<<std::setw(8);
	s<<(unsigned int)(d&0xFFFFFFFF);
	obj<<s.str();
	return obj;
}

int main()
{
	using namespace std;

	//   CRC:

	//  buf:
	// 12 34 56 78 90 AB CD -       .
	uint64_t buf=0xCDAB9078563412;

	//    0xDF
	char crc=crc8((char*)&buf, 7);

	// CRC        CRC  0.
	buf=0xDFCDAB9078563412;		// 12 34 56 78 90 AB CD; DF
	crc=crc8((char*)&buf, 8);	// ==0


	// **************************************************
	//    COM <---> 1-Wire.
	// **************************************************

	Com_1WireBridge bridge;
	bridge.open(COM_NUM, COM_BAUD_RATE);

	//      .
	if(!bridge.reset())
	{
		cout<<"No devices."<<endl;
		return 0;
	}

	//    .
	vector<uint64_t> vrom;
	EnumRom enr(&bridge);
	while(true)
	{
		int b=enr.next();
		if(b>=0)
		{
			vrom.push_back(enr.rom);
			cout<<"ROM: "<<enr.rom<<endl;
		}
		if(b<=0)
			break;
	}

	//     DS18B20,  .
	for(unsigned int i=0; i<vrom.size(); i++)
	{
		if((vrom[i]&0xFF)==0x28)
		{
			if(!bridge.reset())
				continue;
			bridge.send_byte(CMD_MATCH_ROM);
			bridge.send_rom(vrom[i]);
			bridge.send_byte(CMD_CONVERT_T);
			while(!bridge.slot_r())
				Sleep(75);
			bridge.reset();
			bridge.send_byte(CMD_MATCH_ROM);
			bridge.send_rom(vrom[i]);
			bridge.send_byte(CMD_READ_SCRATCHPAD);
			short int fpt=bridge.receive_byte();
			fpt|=bridge.receive_byte()<<8;
			float t=fpt;
			t/=16.0;
			cout<<vrom[i]<<"; t="<<t<<endl;
		}
	}

	cout<<"Press Enter to exit"<<endl;
	cin.get();

	return 0;
}